/*
 * Copyright 2006-2014 The MZmine 2 Development Team
 * 
 * This file is part of MZmine 2.
 * 
 * MZmine 2 is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * MZmine 2 is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * MZmine 2; if not, write to the Free Software Foundation, Inc., 51 Franklin
 * St, Fifth Floor, Boston, MA 02110-1301 USA
 */

package net.sf.mzmine.modules.rawdatamethods.peakpicking.manual;

import java.util.TreeMap;

import javax.annotation.Nonnull;

import net.sf.mzmine.datamodel.DataPoint;
import net.sf.mzmine.datamodel.Feature;
import net.sf.mzmine.datamodel.IsotopePattern;
import net.sf.mzmine.datamodel.RawDataFile;
import net.sf.mzmine.datamodel.Scan;
import net.sf.mzmine.util.CollectionUtils;
import net.sf.mzmine.util.MathUtils;
import net.sf.mzmine.util.PeakUtils;
import net.sf.mzmine.util.Range;
import net.sf.mzmine.util.ScanUtils;

/**
 * This class represents a manually picked chromatographic peak.
 */
class ManualPeak implements Feature {

    private RawDataFile dataFile;

    // Raw M/Z, RT, Height and Area
    private double mz, rt, height, area;

    // Boundaries of the peak
    private Range rtRange, mzRange, intensityRange;

    // Map of scan number and data point
    private TreeMap<Integer, DataPoint> dataPointMap;

    // Number of most intense fragment scan
    private int fragmentScan, representativeScan;

    // Isotope pattern. Null by default but can be set later by deisotoping
    // method.
    private IsotopePattern isotopePattern;
    private int charge = 0;

    /**
     * Initializes empty peak for adding data points
     */
    ManualPeak(RawDataFile dataFile) {
	this.dataFile = dataFile;
	dataPointMap = new TreeMap<Integer, DataPoint>();
    }

    /**
     * This peak is always a result of manual peak detection, therefore MANUAL
     */
    public @Nonnull FeatureStatus getFeatureStatus() {
	return FeatureStatus.MANUAL;
    }

    /**
     * This method returns M/Z value of the peak
     */
    public double getMZ() {
	return mz;
    }

    /**
     * This method returns retention time of the peak
     */
    public double getRT() {
	return rt;
    }

    /**
     * This method returns the raw height of the peak
     */
    public double getHeight() {
	return height;
    }

    /**
     * This method returns the raw area of the peak
     */
    public double getArea() {
	return area;
    }

    /**
     * This method returns numbers of scans that contain this peak
     */
    public @Nonnull int[] getScanNumbers() {
	return CollectionUtils.toIntArray(dataPointMap.keySet());
    }

    /**
     * This method returns a representative datapoint of this peak in a given
     * scan
     */
    public DataPoint getDataPoint(int scanNumber) {
	return dataPointMap.get(scanNumber);
    }

    public @Nonnull Range getRawDataPointsIntensityRange() {
	return intensityRange;
    }

    public @Nonnull Range getRawDataPointsMZRange() {
	return mzRange;
    }

    public @Nonnull Range getRawDataPointsRTRange() {
	return rtRange;
    }

    /**
     * @see net.sf.mzmine.datamodel.Feature#getDataFile()
     */
    public @Nonnull RawDataFile getDataFile() {
	return dataFile;
    }

    public String getName() {
	return PeakUtils.peakToString(this);
    }

    public IsotopePattern getIsotopePattern() {
	return isotopePattern;
    }

    public void setIsotopePattern(@Nonnull IsotopePattern isotopePattern) {
	this.isotopePattern = isotopePattern;
    }

    /**
     * Adds a new data point to this peak
     * 
     * @param scanNumber
     * @param dataPoints
     * @param rawDataPoints
     */
    void addDatapoint(int scanNumber, DataPoint dataPoint) {

	double rt = dataFile.getScan(scanNumber).getRetentionTime();

	if (dataPointMap.isEmpty()) {
	    rtRange = new Range(rt);
	    mzRange = new Range(dataPoint.getMZ());
	    intensityRange = new Range(dataPoint.getIntensity());
	} else {
	    rtRange.extendRange(rt);
	    mzRange.extendRange(dataPoint.getMZ());
	    intensityRange.extendRange(dataPoint.getIntensity());
	}

	dataPointMap.put(scanNumber, dataPoint);

    }

    void finalizePeak() {

	// Trim the zero-intensity data points from the beginning and end
	while (!dataPointMap.isEmpty()) {
	    int scanNumber = dataPointMap.firstKey();
	    if (dataPointMap.get(scanNumber).getIntensity() > 0)
		break;
	    dataPointMap.remove(scanNumber);
	}
	while (!dataPointMap.isEmpty()) {
	    int scanNumber = dataPointMap.lastKey();
	    if (dataPointMap.get(scanNumber).getIntensity() > 0)
		break;
	    dataPointMap.remove(scanNumber);
	}

	// Check if we have any data points
	if (dataPointMap.isEmpty()) {
	    throw (new IllegalStateException(
		    "Peak can not be finalized without any data points"));
	}

	// Get all scan numbers
	int allScanNumbers[] = CollectionUtils
		.toIntArray(dataPointMap.keySet());

	// Find the data point with top intensity and use its RT and height
	for (int i = 0; i < allScanNumbers.length; i++) {
	    DataPoint dataPoint = dataPointMap.get(allScanNumbers[i]);
	    double rt = dataFile.getScan(allScanNumbers[i]).getRetentionTime();
	    if (dataPoint.getIntensity() > height) {
		height = dataPoint.getIntensity();
		representativeScan = allScanNumbers[i];
		this.rt = rt;
	    }
	}

	// Calculate peak area
	area = 0;
	for (int i = 1; i < allScanNumbers.length; i++) {

	    // For area calculation, we use retention time in seconds
	    double previousRT = dataFile.getScan(allScanNumbers[i - 1])
		    .getRetentionTime() * 60d;
	    double currentRT = dataFile.getScan(allScanNumbers[i])
		    .getRetentionTime() * 60d;

	    double rtDifference = currentRT - previousRT;

	    // Intensity at the beginning and end of the interval
	    double previousIntensity = dataPointMap.get(allScanNumbers[i - 1])
		    .getIntensity();
	    double thisIntensity = dataPointMap.get(allScanNumbers[i])
		    .getIntensity();
	    double averageIntensity = (previousIntensity + thisIntensity) / 2;

	    // Calculate area of the interval
	    area += (rtDifference * averageIntensity);

	}

	// Calculate median MZ
	double mzArray[] = new double[allScanNumbers.length];
	for (int i = 0; i < allScanNumbers.length; i++) {
	    mzArray[i] = dataPointMap.get(allScanNumbers[i]).getMZ();
	}
	this.mz = MathUtils.calcQuantile(mzArray, 0.5f);

	fragmentScan = ScanUtils.findBestFragmentScan(dataFile, rtRange,
		mzRange);

	if (fragmentScan > 0) {
	    Scan fragmentScanObject = dataFile.getScan(fragmentScan);
	    int precursorCharge = fragmentScanObject.getPrecursorCharge();
	    if ((precursorCharge > 0) && (this.charge == 0))
		this.charge = precursorCharge;
	}

    }

    public int getRepresentativeScanNumber() {
	return representativeScan;
    }

    public int getMostIntenseFragmentScanNumber() {
	return fragmentScan;
    }

    public int getCharge() {
	return charge;
    }

    public void setCharge(int charge) {
	this.charge = charge;
    }
}
